/**	+deprecated("Please use dub.packagesuppliers")
	Contains (remote) package supplier interface and implementations.	+public import dub.packagesuppliers;

	Copyright: © 2012-2013 Matthias Dondorff, 2012-2016 Sönke Ludwig
	License: Subject to the terms of the MIT license, as written in the included LICENSE.txt file.
	Authors: Matthias Dondorff
*/
deprecated("Please use dub.packagesuppliers")
module dub.packagesupplier;
public import dub.packagesuppliers;
