//          Copyright Ferdinand Majerech 2014.
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

module dub.internal.dyaml.encoding;


import dub.internal.tinyendian;

alias Encoding = dub.internal.tinyendian.UTFEncoding;
